import datetime
import time

import logging
import pytest

from django.contrib.auth import get_user_model
from django.utils.timezone import make_aware
from rest_framework.authtoken.admin import User
from rest_framework.test import APIClient


from greg.models import (
    Invitation,
    InvitationLink,
    OrganizationalUnit,
    Person,
    Role,
    RoleType,
    Sponsor,
)

from gregui.models import GregUserProfile

# faker spams the logs with localisation warnings
# see https://github.com/joke2k/faker/issues/753
logging.getLogger("faker").setLevel(logging.ERROR)

# OIDC stuff
@pytest.fixture
def claims():
    return {
        "sub": "subsub",
        "connect-userid_sec": ["feide:frank_foreleser@spusers.feide.no"],
        "dataporten-userid_sec": [
            # "feide:frank_foreleser@spusers.feide.no"
        ],
        "name": "Frank Foreleser Føllesen",
        "email": "noreply@feide.no",
        "email_verified": True,
        "picture": "https://api.dataporten.no/userinfo/v1/user/media/p:2192dff7-6989-4244-83cc-ae5e78875bdd",
    }


@pytest.fixture
def id_token_payload():
    return {
        "iss": "https://auth.dataporten.no",
        "jti": "jtijti",
        "aud": "lalalalala",
        "sub": "subsub",
        "iat": 1605174731,
        "exp": 1605178331,
        "auth_time": 1605174731,
        "nonce": "noncenonce",
    }


@pytest.fixture
def data():
    return {
        "User": {
            "user1": {
                "username": "user1",
                "email": "user1@example.com",
            },
            "user2": {
                "username": "user2",
                "email": "user2@example.com",
            },
            "https://auth.dataporten.nosubsub": {
                "username": "https://auth.dataporten.nosubsub",
            },
        },
        "Identity": {
            "person1": {
                "type": "feide_id",
                "value": "foo@example.com",
                "person": {"last_name": "Baresen"},
            },
        },
        "Person": {
            "person1": {
                "first_name": "Foo",
                "last_name": "Baresen",
            },
        },
        "Sponsor": {
            "sponsor1": {
                "first_name": "Bar",
                "last_name": "Bazesen",
                "feide_id": "bar@example.com",
            },
        },
    }


def save_object(model, **kwargs):
    obj = model(**kwargs)
    obj.save()
    return obj


# TODO add person and sponsor
OBJECT_MAPPING = {
    "greg" "person": Person,
    "sponsor": Sponsor,
    "user": get_user_model(),
}


def create_objects(cls, data):
    objects = {}
    for obj_name, kwargs in data.items():
        create_kwargs = kwargs.copy()
        for name, selector in kwargs.items():
            if name in OBJECT_MAPPING:
                create_kwargs[name] = OBJECT_MAPPING[name].objects.get(**selector)
        obj = save_object(cls, **create_kwargs)
        objects[obj_name] = obj
    return objects


@pytest.fixture
def client() -> APIClient:
    client = APIClient()
    return client


@pytest.fixture
def unit_foo() -> OrganizationalUnit:
    ou = OrganizationalUnit.objects.create(
        orgreg_id="12345", name_en="Foo EN", name_nb="Foo NB"
    )
    return OrganizationalUnit.objects.get(id=ou.id)


@pytest.fixture
def role_type_foo() -> RoleType:
    rt = RoleType.objects.create(
        identifier="role_foo", name_en="Role Foo EN", name_nb="Role Foo NB"
    )
    return RoleType.objects.get(id=rt.id)


@pytest.fixture
def sponsor_guy_data() -> dict:
    return dict(feide_id="guy@example.org", first_name="Sponsor", last_name="Guy")


@pytest.fixture
def sponsor_guy(unit_foo: OrganizationalUnit, sponsor_guy_data) -> Sponsor:
    sponsor = Sponsor.objects.create(**sponsor_guy_data)
    sponsor.units.add(unit_foo, through_defaults={"hierarchical_access": False})
    return Sponsor.objects.get(id=sponsor.id)


@pytest.fixture
def user_sponsor(sponsor_guy: Sponsor) -> User:
    user_model = get_user_model()

    # Create a user and link him to a sponsor
    user = user_model.objects.create(
        username="test_sponsor",
        email="test@example.org",
        first_name="Test",
        last_name="Sponsor",
    )
    GregUserProfile.objects.create(user=user, sponsor=sponsor_guy)

    # This user is a sponsor for unit_foo
    return user


@pytest.fixture
def person_foo_data() -> dict:
    return dict(
        first_name="Foo",
        last_name="Bar",
    )


@pytest.fixture
def person(person_foo_data) -> Person:
    pe = Person.objects.create(**person_foo_data)
    return Person.objects.get(id=pe.id)


@pytest.fixture
def role(person, sponsor_guy, unit_foo, role_type_foo) -> Role:
    role = Role.objects.create(
        person=person,
        sponsor_id=sponsor_guy,
        orgunit_id=unit_foo,
        end_date="2050-10-15",
        type_id=role_type_foo.id,
    )
    return Role.objects.get(id=role.id)


@pytest.fixture
def invitation(role) -> Invitation:
    inv = Invitation.objects.create(role=role)
    return Invitation.objects.get(id=inv.id)


@pytest.fixture
def invitation_valid_date() -> datetime.datetime:
    return make_aware(datetime.datetime(2060, 10, 15))


@pytest.fixture
def invitation_expired_date() -> datetime.datetime:
    return make_aware(datetime.datetime(1970, 1, 1))


@pytest.fixture
def invitation_link(invitation, invitation_valid_date) -> InvitationLink:
    il = InvitationLink.objects.create(
        invitation=invitation, expire=invitation_valid_date
    )
    return InvitationLink.objects.get(id=il.id)


@pytest.fixture
def invitation_link_expired(invitation, invitation_expired_date) -> InvitationLink:
    il = InvitationLink.objects.create(
        invitation=invitation, expire=invitation_expired_date
    )
    return InvitationLink.objects.get(id=il.id)


@pytest.fixture
def greg_users(data):
    return create_objects(get_user_model(), data["User"])


@pytest.fixture
def greg_persons(data):
    return create_objects(get_user_model(), data["Person"])


@pytest.fixture
def greg_sponsors(data):
    return create_objects(get_user_model(), data["Sponsor"])


@pytest.fixture
def log_in(client, greg_users):
    def _log_in(username):
        user = greg_users[username]
        client.force_login(user=user)
        # It seems like the session was not updated automatically this way
        session = client.session
        session["oidc_id_token_payload"] = {"iat": time.time()}
        session.save()
        return client

    return _log_in
